/*
     This file is part of GNUnet.
     (C) 2001, 2002, 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/afs/protocol/routing.c
 * @brief routing of AFS queries and replies
 * @author Christian Grothoff
 *
 * The routing code is responsible for deciding which replies
 * need to be forwarded to which peers. While the querymanager
 * decides where to forward queries, it needs to negotiate with
 * the routing code which queries can be forwarded since we may
 * not be able to keep track of all queries.
 **/

#include "config.h"

#include <string.h>

#include "util/checksum.h"
#include "util/configuration.h"
#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/statistics.h"
#include "util/statuscalls.h"
#include "util/xmalloc.h"
#include "applications/afs/database/manager.h"
#include "applications/afs/util/policy.h"
#include "applications/afs/util/bloomfilter.h"
#include "applications/afs/database/bloomfilter.h"
#include "applications/afs/protocol/routing.h"
#include "applications/afs/protocol/querymanager.h"
#include "applications/afs/protocol/afs.h"

#define DEBUG_ROUTING NO

/** 
 * Size of the indirection table specified in gnunet.conf
 **/
static unsigned int indirectionTableSize;

static Mutex lock;

/* ****************** config ************** */

/**
 * minimum indirection table size, defaults to 8192 entries, reduce if 
 * you have very little memory, enlarge if you start to overflow often 
 * and have memory available.<p>
 *  
 * If the average query lives for say 1 minute (10 hops), and you have
 * a 56k connection (= 420 kb/minute, or approximately 8000
 * queries/minute) the maximum reasonable routing table size would
 * thus be 8192 entries.  Every entry takes about 56 bytes.<p>
 *  
 * The larger the value is that you pick here, the greater your
 * anonymity can become. It also can improve your download speed.<p>
 *
 * Memory consumption: 
 * <ul>
 * <li>8192 => 450k indirection table => approx. 4 MB gnunetd</li>
 * <li>65536 => 3580k indirection table => approx. 8 MB gnuentd</li>
 * </ul>
 * <p>
 * THE VALUE YOU PICK MUST BE A POWER OF 2, for example:
 * 128, 256, 512, 1024, 2048, 4092, 8192, 16384, 32768, 65536
 **/
/* #define MIN_INDIRECTION_TABLE_SIZE 8192 */
#define MIN_INDIRECTION_TABLE_SIZE 8 


/**
 * Indirection table entry. Lists what we're looking for,
 * where to forward it, and how long to keep looking for it.
 **/
typedef struct {
  /**
   * what are we waiting for? 
   **/
  HashCode160 hash;  

  /**
   * Are we limited to a specific namespace?
   * Non-NULL if yes.
   **/
  HashCode160 * namespace;

  /**
   * when can we forget about this entry? 
   **/
  cron_t ttl;

  /**
   * how much is this query worth? we may want to be a bit more
   * precise here: the query may be worth X for giving credit to
   * somebody who sends a reply to us and Y for forwarding to the
   * originator; notice that the priority already changes after the
   * first reply is received. 
   **/
  unsigned int priority;

  /**
   * which replies have we already seen? 
   **/
  unsigned int seenIndex; 

  /**
   * hashcodes of the encrypted (!) replies that we have forwarded so far 
   **/
  HashCode160 * seen; 

  /**
   * How many hosts are waiting for an answer to this query (length of
   * destination array)
   **/
  unsigned int hostsWaiting; 

  /**
   * who are these hosts? 
   **/
  HostIdentity * destination; 

  /**
   * How many tcpsocks are in use?
   **/
  unsigned int tcpsocksSize;

  /**
   * local TCP clients to send the reply to, NULL if none 
   **/
  GNUNET_TCP_SOCKET ** tcpsocks; 

} IndirectionTableEntry;

/**
 * The routing table. This table has entries for all
 * queries that we have recently send out. It helps 
 * GNUnet to route the replies back to the respective
 * sender.
 **/
static IndirectionTableEntry * ROUTING_indTable_;

/**
 * Stats handle for how much content replies we have
 * send back to clients.
 **/
static int stat_cs_reply_content_out;

/**
 * Stats handles about incoming blocks
 **/
static int stat_content_in_ok;
static int stat_content_in_dupe;
static int stat_content_in_orphan;
static int stat_routingFull;
static int stat_routingReplaced;
static int stat_routingPresent;
static int stat_p2p_query_out;

/**
 * Compute the hashtable index of a host id.
 **/
static unsigned int computeRoutingIndex(HashCode160 * query) {
  unsigned int res = (((unsigned int)query->a) & 
		      ((unsigned int)(indirectionTableSize - 1)));
  if (res >= indirectionTableSize)
    errexit("FATAL: indirectionTableSize not power of 2? (%d)\n",
	    indirectionTableSize);
  return res;
}

/**
 * ITE modes for addToSlot.
 **/
#define ITE_REPLACE 0
#define ITE_GROW 1

/**
 * Call useContent "later" and then free the pmsg.
 **/
static void useSBLOCKContentLater(AFS_p2p_SBLOCK_RESULT * pmsg) {
  useContent(NULL,
	     &pmsg->result.identifier,
	     &pmsg->header);
  FREE(pmsg);
}

/**
 * Call useContent "later" and then free the
 * pmsg.
 **/
static void useCHKContentLater(AFS_p2p_CHK_RESULT * pmsg) {
  HashCode160 hc;

  hash(&pmsg->result,
       sizeof(CONTENT_Block),
       &hc);
  useContent(NULL,
	     &hc,
	     &pmsg->header);
  FREE(pmsg);
}

/**
 * Call useContent "later" and then free the
 * pmsg.
 **/
static void use3HashContentLater(AFS_p2p_3HASH_RESULT * pmsg) {
  HashCode160 hc;

  hash(&pmsg->hash,
       sizeof(HashCode160),
       &hc);
  useContent(NULL,
	     &hc,
	     &pmsg->header);
  FREE(pmsg);
}

/**
 * Queue a CHK reply with cron to simulate
 * another peer returning the response with
 * some latency (and then route as usual).
 * 
 * @param sender the next hop
 * @param result the content that was found
 **/
static void queueCHKReply(HostIdentity * sender,
			  CONTENT_Block * result) {
  AFS_p2p_CHK_RESULT * pmsg;

  pmsg = MALLOC(sizeof(AFS_p2p_CHK_RESULT));
  pmsg->header.size 
    = htons(sizeof(AFS_p2p_CHK_RESULT));
  pmsg->header.requestType 
    = htons(AFS_p2p_PROTO_CHK_RESULT);
  memcpy(&pmsg->result,
	 result,
	 sizeof(CONTENT_Block));
  /* delay reply, delay longer if we are busy (makes it harder
     to predict / analyze, too). */
  addCronJob((CronJob)&useCHKContentLater,
	     randomi(TTL_DECREMENT+cpuUsage()), 
	     0,
	     pmsg);
}

/**
 * Queue an SBLOCK reply with cron to simulate
 * another peer returning the response with
 * some latency (and then route as usual).
 * 
 * @param sender the next hop
 * @param result the content that was found
 **/
static void queueSBLOCKReply(HostIdentity * sender,
			     SBlock * result) {
  AFS_p2p_SBLOCK_RESULT * pmsg;

  pmsg = MALLOC(sizeof(AFS_p2p_SBLOCK_RESULT));
  pmsg->header.size 
    = htons(sizeof(AFS_p2p_SBLOCK_RESULT));
  pmsg->header.requestType 
    = htons(AFS_p2p_PROTO_SBLOCK_RESULT);
  memcpy(&pmsg->result,
	 result,
	 sizeof(CONTENT_Block));
  /* delay reply, delay longer if we are busy (makes it harder
     to predict / analyze, too). */
  addCronJob((CronJob)&useSBLOCKContentLater, 
	     randomi(TTL_DECREMENT+cpuUsage()), 
	     0,
	     pmsg);
}

/**
 * Queue a 3Hash reply with cron to simulate
 * another peer returning the response with
 * some latency (and then route as usual).
 * 
 * @param sender the next hop
 * @param hc the double (!) hash
 * @param result the content that was found
 **/
static void queue3HashReply(HostIdentity * sender,
			    HashCode160 * hc,
			    CONTENT_Block * result) {
  AFS_p2p_3HASH_RESULT * pmsg;

  pmsg = MALLOC(sizeof(AFS_p2p_3HASH_RESULT));
  pmsg->header.size 
    = htons(sizeof(AFS_p2p_3HASH_RESULT));
  pmsg->header.requestType
    = htons(AFS_p2p_PROTO_3HASH_RESULT);
  memcpy(&pmsg->result,
	 result,
	 sizeof(CONTENT_Block));
  memcpy(&pmsg->hash,
	 hc,
	 sizeof(HashCode160));
  /* delay reply, delay longer if we are busy (makes it harder
     to predict / analyze, too). */
  addCronJob((CronJob)&use3HashContentLater,
	     randomi(TTL_DECREMENT+cpuUsage()), 
	     0,
	     pmsg);
}

/**
 * Hand a CHK reply to the client.
 * @param sock the client socket
 * @param result the response
 **/
static void tellClientCHKReply(GNUNET_TCP_SOCKET * sock,
			       CONTENT_Block * result) {
  AFS_CS_RESULT_CHK * reply;
  
  reply = MALLOC(sizeof(AFS_CS_RESULT_CHK));
  reply->header.tcpType
    = htons(AFS_CS_PROTO_RESULT_CHK);
  reply->header.size 
    = htons(sizeof(AFS_CS_RESULT_CHK));
  memcpy(&((AFS_CS_RESULT_CHK*)reply)->result,
	 result,
	 sizeof(CONTENT_Block));
  statChange(stat_cs_reply_content_out, 1);
  writeToSocket(sock,
		&reply->header);
  FREE(reply);
}


/**
 * Hand an SBLOCK reply to the client.
 * @param sock the client socket
 * @param hc the double hash
 * @param result the response
 **/
static void tellClientSBLOCKReply(GNUNET_TCP_SOCKET * sock,
				  SBlock * result) {
  AFS_CS_RESULT_SBLOCK * reply;
  
  reply = MALLOC(sizeof(AFS_CS_RESULT_SBLOCK));
  reply->header.tcpType
    = htons(AFS_CS_PROTO_RESULT_SBLOCK);
  reply->header.size 
    = htons(sizeof(AFS_CS_RESULT_SBLOCK));
  memcpy(&reply->result,
	 result,
	 sizeof(CONTENT_Block));
  statChange(stat_cs_reply_content_out, 1);
  writeToSocket(sock,
		&reply->header);
  FREE(reply);
}

/**
 * Hand a 3Hash reply to the client.
 * @param sock the client socket
 * @param hc the double hash
 * @param result the response
 **/
static void tellClient3HashReply(GNUNET_TCP_SOCKET * sock,
				 HashCode160 * hc,
				 CONTENT_Block * result) {
  AFS_CS_RESULT_3HASH * reply;
  
  reply = MALLOC(sizeof(AFS_CS_RESULT_3HASH));
  reply->header.tcpType
    = htons(AFS_CS_PROTO_RESULT_3HASH);
  reply->header.size 
    = htons(sizeof(AFS_CS_RESULT_3HASH));
  memcpy(&reply->result,
	 result,
	 sizeof(RootNode));
  memcpy(&reply->hash,
	 hc,
	 sizeof(HashCode160));    
  statChange(stat_cs_reply_content_out, 1);
  writeToSocket(sock,
		&reply->header);
  FREE(reply);
}

/**
 * Add an entry to the routing table. The lock on the ite
 * must be held and is being released.
 *
 * @param mode replace or extend an existing entry?
 * @param ite slot in the routing table that is manipulated
 * @param query the query to look for
 * @param ttl how long to keep the new entry, relative ttl
 * @param priority how important is the new entry
 * @param sender for which node is the entry (NULL for local client)
 * @param sock for which local client is the entry (NULL for peer)
 **/
static void addToSlot(int mode,
		      IndirectionTableEntry * ite,
		      HashCode160 * query,
		      HashCode160 * namespace,
		      int ttl,
		      unsigned int priority,
		      HostIdentity * sender,
		      GNUNET_TCP_SOCKET * sock) {
  unsigned int i;
  cron_t now;

  /* namespace handling: always override with 
     the new value (query collisions are 
     supposed to be 'impossible', so this should
     always be correct.  Either we replace the
     existing slot with something new, or it
     should not make a difference since the old 
     and the new namespace will be the same. */
  if (ite->namespace != NULL) {
    if (namespace == NULL) {
      FREE(ite->namespace);
      ite->namespace = NULL;
    } else {
      memcpy(ite->namespace,
	     namespace,
	     sizeof(HashCode160));
    }
  } else {
    if (namespace != NULL) {
      ite->namespace = MALLOC(sizeof(HashCode160));
      memcpy(ite->namespace,
	     namespace,
	     sizeof(HashCode160));
    }
  }
  cronTime(&now);
  if (mode == ITE_REPLACE) {
    if (equalsHashCode160(query,
			  &ite->hash)) {
      statChange(stat_routingPresent, 1); 
      ite->ttl = now + ttl;
      ite->priority = priority;
    } else {
      statChange(stat_routingReplaced, 1);
      GROW(ite->seen,
	   ite->seenIndex,
	   0);
      memcpy(&ite->hash,
	     query,
	     sizeof(HashCode160));
      GROW(ite->destination,
	   ite->hostsWaiting,
	   0);
      GROW(ite->tcpsocks,
	   ite->tcpsocksSize,
	   0);
      ite->ttl = now + ttl;
      ite->priority = priority;      
    }
  } else {
    if (!equalsHashCode160(query,
			   &ite->hash))
      errexit("FATAL: assert failed in %s.%d\n", 
	      __FILE__, 
	      __LINE__);
    if (sender != NULL) 
      for (i=0;i<ite->hostsWaiting;i++)
	if (equalsHashCode160(&sender->hashPubKey,
			      &ite->destination[i].hashPubKey)) {
	  sender = NULL;
	  break;
	}    
    statChange(stat_routingPresent, 1);
    if (sock != NULL) 
      for (i=0;i<ite->tcpsocksSize;i++)
	if (sock == ite->tcpsocks[i]) {
	  sock = NULL;
	  break;
	}
    if ( (sock == NULL) && 
	 (sender == NULL) ) {
      MUTEX_UNLOCK(&lock); 
      return; /* already there! */  
    }
    /* extend lifetime */
    if (ite->ttl < now + ttl)
      ite->ttl = now + ttl; 
    ite->priority += priority;
  }
  if (sock != NULL) { 
    for (i=0;i<ite->tcpsocksSize;i++)
      if (ite->tcpsocks[i] == sock) 
	sock = NULL;
    if (sock != NULL) {
      GROW(ite->tcpsocks,
	   ite->tcpsocksSize,
	   ite->tcpsocksSize+1);
      ite->tcpsocks[ite->tcpsocksSize-1] = sock;
      GROW(ite->seen,
	   ite->seenIndex,
	   0); /* new listener, flush "seen" list */
    }
  }
  if (sender != NULL) {
    for (i=0;i<ite->hostsWaiting;i++)
      if (equalsHashCode160(&ite->destination[i].hashPubKey,
			    &sender->hashPubKey)) {
	sender = NULL;
	break;
      }
  }
  if (sender != NULL) {
    GROW(ite->destination,
	 ite->hostsWaiting,
	 ite->hostsWaiting+1);
    memcpy(&ite->destination[ite->hostsWaiting-1],
	   sender,
	   sizeof(HostIdentity));
    /* again: new listener, flush seen list */
    GROW(ite->seen,
	 ite->seenIndex,
	 0);
  }
  MUTEX_UNLOCK(&lock); 
}

/**
 * Find out, if this query is already pending. If the ttl of
 * the new query is higher than the ttl of an existing query,
 * NO is returned since we should re-send the query.<p>
 *
 * If YES is returned, the slot is also marked as used by
 * the query and the sender (HostId or socket) is added.<p>
 *
 * This method contains a heuristic that attempts to do its best to
 * route queries without getting too many cycles, send a query and
 * then drop it from the routing table without sending a response,
 * etc.  Before touching this code, definitely consult Christian
 * (grothoff@cs.purdue.edu) who has put more bugs in these five lines
 * of code than anyone on this planet would think is possible.
 *
 *
 * @param query the hash to look for
 * @param ttl how long would the new query last 
 * @param priority the priority of the query
 * @param sender which peer transmitted the query? (NULL for this peer)
 * @param sock which client transmitted the query? (NULL for other peer)
 * @param isRouted set to OK if we can route this query, SYSERR if we can not
 * @param doForward is set to OK if we should forward the query, SYSERR if not
 **/
static void needsForwarding(HashCode160 * query,
			    HashCode160 * namespace,
			    int ttl,
			    unsigned int priority,
			    HostIdentity * sender,
			    GNUNET_TCP_SOCKET * sock,
			    int * isRouted,
			    int * doForward) {
  IndirectionTableEntry * ite;
  cron_t now;

  cronTime(&now);
  ite = &ROUTING_indTable_[computeRoutingIndex(query)];
  MUTEX_LOCK(&lock); 

  if ( (ttl > 0) &&
       (ttl > (int)priority*1000) )
    ttl = (int) priority*1000; /* bound! */
  if ( ( ttl < 0) &&
       (equalsHashCode160(query,
			  &ite->hash) ) ) {
    /* if ttl is "expired" and we have
       the exact query pending, route 
       replies but do NOT forward _again_! */
#if DEBUG_ROUTING
    LOG(LOG_INFO,
	"INFO: GROW: ttl < 0 and existing query is equal (%d, %d)\n",
	ttl, 
	(int) (ite->ttl - now));
#endif
    addToSlot(ITE_GROW, ite, query, namespace, ttl, priority, sender, sock);
    *isRouted = YES;
    *doForward = NO;
    return; 
  }
 
  if ( (ite->ttl + (TTL_DECREMENT * coreAPI->estimateNetworkSize()) < 
	(cron_t)(now + ttl)) &&
       (ite->ttl < now) ) { 
    /* expired AND is significantly (!) 
       longer expired than new query */
#if DEBUG_ROUTING
    LOG(LOG_INFO,
	"INFO: REPLACE and reset SEEN: existing query "
	"expired and older than new query (%d, %d)\n",
	ttl, 
	(int) (ite->ttl - now));
#endif
    /* previous entry relatively expired, start using the slot --
       and kill the old seen list!*/
    GROW(ite->seen,
	 ite->seenIndex,
	 0);
    addToSlot(ITE_REPLACE, ite, query, namespace, ttl, priority, sender, sock);
    *isRouted = YES;
    *doForward = YES;
    return;
  }
  if (equalsHashCode160(query,
			&ite->hash) ) {
    /* pending == new! */
    if ( (ite->seenIndex > 0) &&
	 (equalsHashCode160(&ite->hash,
			    &ite->seen[0])) ) { /* CHK */
      if ( (ite->seenIndex == 1) &&
	   (ite->ttl < (cron_t)(now + ttl)) ) { /* reply seen & ttl of new is longer?*/
	/* We have seen a reply to this CHK recently before,
	   BUT the new TTL is higher than the old one, so
	   go again */
	GROW(ite->seen,
	     ite->seenIndex,
	     0);
#if DEBUG_ROUTING
	LOG(LOG_INFO,
	    "INFO: REPLACE and reset SEEN: existing query equal "
	    "but we've seen the response already (%d, %d)\n",
	    (int) (ite->ttl - now),
	    ttl);
#endif
	addToSlot(ITE_REPLACE, ite, query, namespace, ttl, priority, sender, sock);
	*isRouted = YES;
	*doForward = YES;
	return;
      } else {
	/* We have not seen a reply for this CHK before,
	   or the new TTL is lower than the old one, thus
	   just wait for the reply that may come back */
#if DEBUG_ROUTING
	LOG(LOG_INFO,
	    "INFO: GROW - equal existing query exists without replies (%d, %d)\n",
	    (int) (ite->ttl - now),
	    ttl);	    
#endif
	addToSlot(ITE_GROW, ite, query, namespace, ttl, priority, sender, sock);
	*isRouted = YES;
	*doForward = NO;
	return;
      }
    } else { /* 3HASH, multiple results possible! */
      /* It's a pending 3HASH that can have multiple 
	 replies.  Do not re-send, just forward the
	 answers that we get from now on to this additional
	 receiver */
#if DEBUG_ROUTING
      LOG(LOG_INFO,
	  "INFO: GROW - equal existing query exists without replies (%d, %d)\n",
	  (int) (ite->ttl - now), 
	  ttl);	    
#endif
      addToSlot(ITE_GROW, ite, query, namespace, ttl, priority, sender, sock);
      *isRouted = YES;
      *doForward = NO;
      return;
    }
  } 
  /* a different query that is expired a bit longer is using
     the slot; but if it is a CHK query that has received
     a response already, we can eagerly throw it out 
     anyway, since the request has been satisfied 
     completely */
  if ( (ite->ttl + TTL_DECREMENT < (cron_t)(now + ttl) ) &&
       (ite->ttl < now) && 
       (ite->seenIndex == 1) &&
       (equalsHashCode160(&ite->hash,
			  &ite->seen[0])) ) {
    /* is CHK and we have seen the answer, get rid of it early */
#if DEBUG_ROUTING
    HexName old;
    IFLOG(LOG_INFO,
	  hash2hex(&ite->hash,
		   &old));
    LOG(LOG_INFO,
	"INFO: CHK %s with reply already seen, replacing eagerly (%d, %d).\n",
	&old,
	(int) (ite->ttl - now),
	ttl);	    
#endif
    addToSlot(ITE_REPLACE, ite, query, namespace, ttl, priority, sender, sock);
    *doForward = YES;
    *isRouted = YES;
    return;
  }   
  /* Another still valid query is using the slot.  Now we need a _really_
     good reason to discard it... */
  if (ttl < 0) {
    MUTEX_UNLOCK(&lock);
    *doForward = NO;
    *isRouted = NO;
    return; /* if new ttl is "expired", don't bother with priorities */
  }

  /* Finally try to find a _strong_ reason looking at priority/ttl
     relationships to replace the existing query. A low ttl with high
     priority should be preferred, so we do a cross-multiplication
     (!). Also, we want a _strong_ reason, so we add a "magic" factor
     of 10 for the additional work that the replacement would make
     (the network needs a certain amount of resilience to changes in
     the routing table, otherwise it might happen that query A
     replaces query B which replaces query A which could happen so
     quickly that no response to either query ever makes it through...
  */
  if ( (long long)((ite->ttl - now) * priority) > (long long) 10 * (ttl * ite->priority)) {
#if DEBUG_ROUTING
    LOG(LOG_INFO,
	"INFO: priority of new query is much higher, overriding (%d, %d).\n",
	(int) (ite->ttl - now), 
	ttl);	    
#endif
    addToSlot(ITE_REPLACE, ite, query, namespace, ttl, priority, sender, sock);
    *doForward = YES;
    *isRouted = YES;
    return;
  }  
  /* sadly, the slot is busy with something else; we can 
     not even add ourselves to the reply set */
  statChange(stat_routingFull, 1);
  MUTEX_UNLOCK(&lock); 
  *doForward = NO;
  *isRouted = NO;
  return;
}

/**
 * Send a reply to a host.  Distinguishes between local and remote
 * delivery, converts the reply into the appropriate format and sends
 * it out.
 *
 * @param ite the matching slot in the indirection table
 * @param msg the message to route
 **/
static void sendReply(IndirectionTableEntry * ite,
		      p2p_HEADER * msg) {
  unsigned int j;
   
  /* send to peers */  
  for (j=0;j<ite->hostsWaiting;j++)
    coreAPI->sendToNode(&ite->destination[j],
			msg, 
			ite->priority*2*sizeof(CONTENT_Block)/sizeof(HashCode160), 
			/* weigh priority (twice!) to avoid prefering sending
			   queries over replies! */
			TTL_DECREMENT);    
  for (j=0;j<ite->tcpsocksSize;j++) {
    switch (ntohs(msg->requestType)) {
    case AFS_p2p_PROTO_3HASH_RESULT:
      tellClient3HashReply(ite->tcpsocks[j],
			   &((AFS_p2p_3HASH_RESULT*)msg)->hash,
			   (CONTENT_Block*)&((AFS_p2p_3HASH_RESULT*)msg)->result);
      break;
    case AFS_p2p_PROTO_CHK_RESULT:
      tellClientCHKReply(ite->tcpsocks[j],
			 &((AFS_p2p_CHK_RESULT*)msg)->result);
      break;
    case AFS_p2p_PROTO_SBLOCK_RESULT:
      tellClientSBLOCKReply(ite->tcpsocks[j],
			    &((AFS_p2p_SBLOCK_RESULT*)msg)->result);
      break;
    default:
      LOG(LOG_WARNING,
	  "WARNING: p2p result has unexpected type: %d\n",
	  ntohs(msg->requestType));
      break;
    }    
  }
}
 
/**
 * TCP connection is shut down, cancel all replies to that client.
 **/
static void cancelTCP_routing(GNUNET_TCP_SOCKET * sock) {
  unsigned int i;
  unsigned int j;
  IndirectionTableEntry * ite;

  MUTEX_LOCK(&lock);
  for (i=0;i<indirectionTableSize;i++) {
    ite = &ROUTING_indTable_[i];  
    for (j=0;j<ite->tcpsocksSize;j++)
      if (ite->tcpsocks[j] == sock) {
	ite->tcpsocks[j] 
	  = ite->tcpsocks[--ite->tcpsocksSize];
	/* we don't shrink, not worth it */
      }
  }
  MUTEX_UNLOCK(&lock);  
}

/**
 * Execute a single query. Tests if the query can be routed. If yes, 
 * the query is added to the routing table and the content is looked 
 * for locally. If the content is available locally, a deferred 
 * response is simulated with a cron job and the local content is 
 * marked as valueable. The method returns OK if the query should 
 * subsequently be routed to other peers.
 *
 * @param sender next hop in routing of the reply
 * @param sock client socket if we are ultimate receiver
 * @param prio the effective priority of the query
 * @param ttl the relative ttl of the query
 * @param query the query itself
 * @param superHash YES if the super-hash test has indicated that we#
 *        have a reply locally available
 * @return OK if the query should be routed further, SYSERR if not.
 **/
static int execSingleQuery(HostIdentity * sender,
			   GNUNET_TCP_SOCKET * sock,
			   unsigned int prio,
			   int ttl,
			   HashCode160 * query,
			   int superHash) {
  ContentIndex ce;
  CONTENT_Block * result;
  int len;  
  int isRouted;
  int doForward;
  
  needsForwarding(query,
		  NULL,
		  ttl,
		  prio,
		  sender,
		  sock,
		  &isRouted,
		  &doForward); 
  if (isRouted == SYSERR)
    return SYSERR; /* if we can't route, 
		      forwarding never makes
		      any sense */
  if ( (NO == superHash) &&
       (NO == testBloomfilter(singleBloomFilter,
			      query)) ) 
    return doForward; /* content not available locally,
			 just route */

  result = NULL;
  len = retrieveContent(query,
			&ce,
			(void**)&result,
			prio,
			sender == NULL);
  if (len == -1)
    return doForward; /* bloomfilter was wrong, content not there */
  if (sender != NULL) {
    if (ntohs(ce.type) == LOOKUP_TYPE_3HASH) {
      if (NO == checkAnonymityPolicy(AFS_CS_PROTO_RESULT_3HASH,
				     sizeof(AFS_p2p_3HASH_RESULT))) {
	FREENONNULL(result);
	return doForward; /* policy says: no direct response, but
			     routing is ok! */
      }
    } else {
      if (NO == checkAnonymityPolicy(AFS_CS_PROTO_RESULT_CHK,
				     sizeof(AFS_p2p_CHK_RESULT))) {
	FREENONNULL(result);
	return doForward; /* policy says: no direct response, but 
			     routing is ok */
      }
    }
  }
  switch (ntohs(ce.type)) {
  case LOOKUP_TYPE_CHK:
  case LOOKUP_TYPE_CHKS:
    if (len != sizeof(CONTENT_Block)) {
      LOG(LOG_WARNING,
	  "WARNING: local CHK content had bad size %d\n",
	  len);
      break;
    }
    if (sock != NULL) {
      tellClientCHKReply(sock,
			 result);
      doForward = SYSERR; /* purely local handling! */
    }
    if (sender != NULL)
      queueCHKReply(sender,
		    result);
    break;
  case LOOKUP_TYPE_3HASH: {
    int * perm;
    unsigned int i;
    int rcount;
    int tmp;

    rcount = len / sizeof(CONTENT_Block);
    if (rcount * sizeof(CONTENT_Block) != (unsigned int) len) {
      LOG(LOG_WARNING,
	  "WARNING: database returned a 3HASH result "\
	  "that is not multiple of 1k (%d)! Database corrupt?\n",
	  len);
      break;
    }
    if (sock != NULL)
      for (i=0;i<(unsigned int)rcount;i++)
	tellClient3HashReply(sock,
			     &ce.hash,
			     &result[i]);
    if (sender != NULL) {
      perm = permute(rcount);
      tmp = (50-getNetworkLoad())*prio;
      if (tmp < rcount)
	rcount = tmp;
      if (rcount <= 0)
	rcount = 1;
      for (i=0;i<(unsigned int)rcount;i++)
        queue3HashReply(sender,
		        &ce.hash,
		        &result[perm[i]]);
      FREE(perm);
    }
    break;
  }
  default:
    LOG(LOG_DEBUG,
	"DEBUG: query lookup produced unexpected type %d!\n",
	ntohs(ce.type));
    break;
  }
  FREENONNULL(result);
  return doForward;
}


/**
 * Execute a namespace query. Tests if the query can be routed. If yes, 
 * the query is added to the routing table and the content is looked 
 * for locally. If the content is available locally, a deferred 
 * response is simulated with a cron job and the local content is 
 * marked as valueable. The method returns OK if the query should 
 * subsequently be routed to other peers.
 *
 * @param sender next hop in routing of the reply
 * @param sock client socket if we are ultimate receiver
 * @param prio the effective priority of the query
 * @param ttl the relative ttl of the query
 * @param query the query itself
 * @param namespace for which namespace
 * @return OK if the query should be routed further, SYSERR if not.
 **/
static int execNSQuery(HostIdentity * sender,
		       GNUNET_TCP_SOCKET * sock,
		       unsigned int prio,
		       int ttl,
		       HashCode160 * query,
		       HashCode160 * namespace) {
  ContentIndex ce;
  SBlock * result;
  int len;  
  HashCode160 hc;
  HexName hex1;
  HexName hex2;
  int isRouted;
  int doForwarding;
  
  IFLOG(LOG_DEBUG,
	hash2hex(query,
		 &hex1));
  IFLOG(LOG_DEBUG,
	hash2hex(namespace,
		 &hex2));
  LOG(LOG_DEBUG,
      "DEBUG: received NS query for %s/%s\n",
      &hex2,
      &hex1);
  
  needsForwarding(query,
		  namespace,
		  ttl,
		  prio,
		  sender,
		  sock,
		  &isRouted,
		  &doForwarding);
  if (SYSERR == isRouted)
    return SYSERR;
  if (NO == testBloomfilter(singleBloomFilter,
			    query)) {
    LOG(LOG_DEBUG,
	"DEBUG: bloomfilter test says content not available locally.\n");
    return doForwarding; /* content not available locally,
			    just route */
  }

  result = NULL;
  len = retrieveContent(query,
			&ce,
			(void**)&result,
			prio,
			sender == NULL);
  if (len == -1) {
    LOG(LOG_DEBUG,
	"DEBUG: bloomfilter test wrong, DB lookup failed.\n");
    return doForwarding; /* bloomfilter was wrong, content not there */
  }
  if (ntohs(ce.type) != LOOKUP_TYPE_SBLOCK) {
    FREE(result);
    return doForwarding;
  }
  if (sender != NULL)
    if (NO == checkAnonymityPolicy(AFS_CS_PROTO_RESULT_SBLOCK,
				   sizeof(AFS_p2p_SBLOCK_RESULT))) {
      LOG(LOG_DEBUG,
	  "DEBUG: anonymity policy denies reply at this time.\n");
      FREE(result);
      return doForwarding; /* policy says: no direct response, but 
			      routing is ok */  
    }
  if (len != sizeof(CONTENT_Block)) {
    LOG(LOG_WARNING,
	"WARNING: local SBLOCK content had bad size %d\n",
	len);
    FREE(result);
    return doForwarding;
  }
  hash(&result->subspace,
       sizeof(PublicKey),
       &hc);
  if (! equalsHashCode160(namespace,
			  &hc)) {
    LOG(LOG_WARNING,
	"WARNING: namespace mismatch (should be rare but can theoretically happen).\n");
    FREE(result);
    return doForwarding;
  }
  if (sender != NULL)
    queueSBLOCKReply(sender,
		     result);
  if (sock != NULL) {
    tellClientSBLOCKReply(sock,
			  result);
    doForwarding = SYSERR;
  }
  FREENONNULL(result);
  return doForwarding;
}

 
/* ****************** public methods ****************** */

/**
 * Initialize routing module (initializes indirection table)
 **/
void initRouting() {
  unsigned int i;

  MUTEX_CREATE(&lock);
  stat_cs_reply_content_out 
    = statHandle("# kb downloaded by clients");
  stat_content_in_ok
    = statHandle("# kb ok content in");
  stat_content_in_dupe
    = statHandle("# kb dupe content in");
  stat_content_in_orphan
    = statHandle("# kb orphan or pushed content in");
  stat_routingFull
    = statHandle("# routing table full");
  stat_routingReplaced
    = statHandle("# routing table entry replaced");
  stat_routingPresent
    = statHandle("# routing table entry already in place");
  stat_p2p_query_out
    = statHandle("# p2p queries sent");
  indirectionTableSize =
    getConfigurationInt("AFS",
    			"INDIRECTIONTABLESIZE");
  if (indirectionTableSize < MIN_INDIRECTION_TABLE_SIZE)
    indirectionTableSize = MIN_INDIRECTION_TABLE_SIZE;
  i = 1;
  while (i < indirectionTableSize)
    i*=2;
  indirectionTableSize = i; /* make sure it's a power of 2 */
#if DEBUG_ROUTING
  LOG(LOG_DEBUG,
      "DEBUG: Set indirectiontablesize to %d\n",
      indirectionTableSize);
#endif
  ROUTING_indTable_ 
    = MALLOC(sizeof(IndirectionTableEntry)*
	     indirectionTableSize);
  for (i=0;i<indirectionTableSize;i++) {
    ROUTING_indTable_[i].namespace = NULL;
    ROUTING_indTable_[i].ttl = 0; /* expired / free */  
    ROUTING_indTable_[i].seenIndex = 0;
    ROUTING_indTable_[i].seen = NULL;
    ROUTING_indTable_[i].hostsWaiting = 0; /* expired / free */  
    ROUTING_indTable_[i].destination = NULL;
    ROUTING_indTable_[i].tcpsocksSize = 0;
    ROUTING_indTable_[i].tcpsocks = NULL;
  }
  coreAPI->registerClientExitHandler(&cancelTCP_routing);
}

/**
 * Shutdown the routing module.
 **/
void doneRouting() {
  unsigned int i;

  for (i=0;i<indirectionTableSize;i++) {
    FREENONNULL(ROUTING_indTable_[i].namespace);
    ROUTING_indTable_[i].namespace = NULL;
    GROW(ROUTING_indTable_[i].seen, 
	 ROUTING_indTable_[i].seenIndex, 
	 0);
    GROW(ROUTING_indTable_[i].destination, 
	 ROUTING_indTable_[i].hostsWaiting,
	 0);
    GROW(ROUTING_indTable_[i].tcpsocks,
	 ROUTING_indTable_[i].tcpsocksSize,
	 0);	 
  }
  MUTEX_DESTROY(&lock);      
  coreAPI->unregisterClientExitHandler(&cancelTCP_routing);
  FREE(ROUTING_indTable_);
}

/**
 * Print the routing table.
 **/
void printRoutingTable() {
  unsigned int i;
  IndirectionTableEntry * ite;
  HexName h1;
  cron_t now;

  cronTime(&now);
  LOG(LOG_MESSAGE,
      "Routing TABLE:\n");
  MUTEX_LOCK(&lock);
  for (i=0;i<indirectionTableSize;i++) {
    ite = &ROUTING_indTable_[i];  
    IFLOG(LOG_MESSAGE,
	  hash2hex(&ite->hash, 
		   &h1));
    /* if (ite->ttl >= now)*/
      LOG(LOG_DEBUG,
	  "DEBUG: %u: hash %s ttl %ds "
	  "hostsWaiting %d prio %d seenIndex: %d\n", 
	  i,
	  &h1, 
	  (int) (((long long) ite->ttl - (long long)now)/cronSECONDS),
	  ite->hostsWaiting,
	  ite->priority,
	  ite->seenIndex);
  }
  MUTEX_UNLOCK(&lock);    
}

/**
 * Execute the query. <p>
 *
 * Execute means to test if we can route the query (or, in the case
 * of a multi-query, any of the sub-queries). If yes, we lookup the
 * content locally and potentially route it deferred. Regardless if
 * the content was found or not, the queries that we can route are
 * forwarded to other peers (by the querymanager code).<p>
 *
 * The decision if we can route is made by "needsForwarding". Note that
 * queries that we are already routing do not "need forwarding". If
 * we do route the query, execQuery decides if we are going to do source
 * rewriting or not.<p>
 *
 * If we route a query, execSingleQuery will use the bloom filters and
 * the databases to locate the content and queue a cron job that will
 * pass the response to "useContent" as if it came from another peer.
 * Note that if the query originated from a local client, the response
 * is instant (no cron job scheduled).
 * 
 * @param qp the polciy (priority) for the query
 * @param msg the query message (with host identity for the reply)
 * @param sock the TCP socket to send the answer to if it is
 *        a query from the local host, otherwise NULL.
 **/
void execQuery(QUERY_POLICY qp, 
	       AFS_p2p_QUERY * msg,
	       GNUNET_TCP_SOCKET * sock) {  
  HostIdentity * sender;
#if DEBUG_ROUTING
  HexName queryHex;
#endif
  HostIdentity senderCpy;
  unsigned int prio;
  int count;
  int routeCount;

  count = (ntohs(msg->header.size)-sizeof(AFS_p2p_QUERY)) / sizeof(HashCode160);
  prio = ntohl(msg->priority);
  if ( (qp & QUERY_PRIORITY_BITMASK) < prio) {
    prio = qp & QUERY_PRIORITY_BITMASK;
    msg->priority = htonl(prio);
  }  
  prio = prio / count; 
  /* source rewriting (or not...) */
  if (sock == NULL) {
    if (equalsHashCode160(&msg->returnTo.hashPubKey,
			  &coreAPI->myIdentity->hashPubKey))
      return; /* A to B, B sends back to A without (!) source rewriting,
		 in this case, A must just drop */
    memcpy(&senderCpy,
	   &msg->returnTo,
	   sizeof(HostIdentity));
    sender = &senderCpy;
  } else {
    sender = NULL;
    memcpy(&senderCpy,
	   coreAPI->myIdentity,
	   sizeof(HostIdentity));
  }
  if ((qp & QUERY_INDIRECT) > 0) 
    memcpy(&msg->returnTo,
	   coreAPI->myIdentity,
	   sizeof(HostIdentity));
  else 
    msg->priority = ntohl(0);  


#if DEBUG_ROUTING
  IFLOG(LOG_INFO,
	hash2hex(&msg->queries[0],
		 &queryHex));
  LOG(LOG_INFO,
      "INFO: received %d-query %s with ttl %d and priority %u\n",
      count,
      &queryHex,
      ntohl(msg->ttl),
      ntohl(msg->priority));
#endif
  if (ntohs(msg->header.requestType) == AFS_p2p_PROTO_NSQUERY) {
    if (OK == execNSQuery(sender,
			  sock,
			  prio,
			  ntohl(msg->ttl),
			  &((AFS_p2p_NSQUERY*)msg)->identifier,
			  &((AFS_p2p_NSQUERY*)msg)->namespace))
      routeCount = 2; /* NAMESPACE + IDENTIFIER! */
    else
      routeCount = 0;  
  } else {
    if (count > 1) { /* MUTLI-QUERY, take apart for individual
			routing, but reassemble for forwarding */
      int i;
      int superBF;
      
      superBF = testBloomfilter(superBloomFilter,
				&msg->queries[0]);
      routeCount = 1;
      for (i=1;i<count;i++) {
	if (OK == execSingleQuery(sender,
				  sock,
				  prio,
				  ntohl(msg->ttl),
				  &msg->queries[i],
				  superBF)) {
	  /* route this query! */
	  memcpy(&msg->queries[routeCount],
		 &msg->queries[i],
		 sizeof(HashCode160));
	  routeCount++;
	}
      }
      if (routeCount == 1)
	routeCount = 0; /* nothing to forward */
    } else { /* single query or 3hash search */
      if (OK == execSingleQuery(sender,
				sock,
				prio,
				ntohl(msg->ttl),
				&msg->queries[0],
				NO))
	routeCount = 1;
      else
	routeCount = 0;
    }
  }

  if (routeCount >= 1) {
    statChange(stat_p2p_query_out, routeCount);
    msg->header.size = htons(sizeof(AFS_p2p_QUERY) + 
			     routeCount * sizeof(HashCode160));
    forwardQuery(msg);
#if DEBUG_ROUTING
    LOG(LOG_DEBUG,
	"DEBUG: slots free in routing table, forwarded %d out of %d queries\n",
	routeCount, 
	count);
#endif
  } else {
#if DEBUG_ROUTING
    LOG(LOG_DEBUG,
	"DEBUG: 0 slots free in routing table for query with %d hash codes, none forwarded\n",
	count);
#endif
  }
}

/**
 * Content has arrived. We must decide if we want to a) forward it to
 * our clients b) indirect it to other nodes. The routing module
 * should know what to do.  This method checks the routing table if
 * we have a matching route and if yes queues the reply. It also makes
 * sure that we do not send the same reply back on the same route more
 * than once.
 *
 * @param hostId who sent the content? NULL for locally found content.
 * @param queryHash either the triple hash or the CHK of the content
 * @param msg the p2p reply to send 
 * @return how good this content was (effective priority of the original request)
 **/
int useContent(const HostIdentity * hostId,
	       HashCode160 * queryHash,
	       p2p_HEADER * msg) {
  unsigned int i;
  CONTENT_Block * content;
  HashCode160 contentHC;
  IndirectionTableEntry * ite;
  int prio = -1;
  cron_t now;
  HexName hex;
  HexName peer;
  
  /*LOG(LOG_DEBUG,
      "DEBUG: useContent - prints routing table\n");
      printRoutingTable();*/
  IFLOG(LOG_DEBUG,
	hash2hex(queryHash, &hex));
  if (hostId != NULL)
    hash2hex(&hostId->hashPubKey, &peer);
#if DEBUG_ROUTING
  LOG(LOG_DEBUG, 
      "DEBUG: received content %s from peer %s\n",
      &hex,
      (hostId == NULL) ? "self" : (char*)&peer);
#endif

  cronTime(&now);
  ite = &ROUTING_indTable_[computeRoutingIndex(queryHash)];
  MUTEX_LOCK(&lock);
 
  if (!equalsHashCode160(&ite->hash,
			 queryHash) ) {	
    statChange(stat_content_in_orphan, 1);
    MUTEX_UNLOCK(&lock);
#if DEBUG_ROUTING
    LOG(LOG_DEBUG, 
	"DEBUG: no matching query pending for content (not indirected)\n");
#endif
    return 0; /* no indirection pending: was useless */
  }

  switch (ntohs(msg->requestType)) {
  case AFS_p2p_PROTO_3HASH_RESULT:
    content = (CONTENT_Block*) &((AFS_p2p_3HASH_RESULT*)msg)->result;
    if (ite->namespace != NULL) {
      MUTEX_UNLOCK(&lock);
      return 0;
    }
    break;
  case AFS_p2p_PROTO_CHK_RESULT:
    content = &((AFS_p2p_CHK_RESULT*)msg)->result;
    if (ite->namespace != NULL) {
      MUTEX_UNLOCK(&lock);
      return 0;
    }
    break;
  case AFS_p2p_PROTO_SBLOCK_RESULT:
    content = (CONTENT_Block*) &((AFS_p2p_SBLOCK_RESULT*)msg)->result;
    if (ite->namespace == NULL) {
      MUTEX_UNLOCK(&lock);
      return 0;
    } else {
      HashCode160 hc;
      hash(&((AFS_p2p_SBLOCK_RESULT*)msg)->result.subspace,
	   sizeof(PublicKey),
	   &hc);
      if (! equalsHashCode160(ite->namespace,
			      &hc) ) {
	MUTEX_UNLOCK(&lock);
	return 0;
      }
    }
    break;
  default:
    LOG(LOG_WARNING,
	"WARNING: p2p result has unexpected type: %d\n",
	ntohs(msg->requestType));
    MUTEX_UNLOCK(&lock);
    return 0;
  }
  hash(content,
       sizeof(CONTENT_Block),
       &contentHC);

  for (i=0;i<ite->seenIndex;i++) {
    if (equalsHashCode160(&contentHC,
			  &ite->seen[i])) {
     statChange(stat_content_in_dupe, 1);
     /* LOG(LOG_DEBUG, 
	"DEBUG: content is not new.\n"); */
     MUTEX_UNLOCK(&lock);
     return 0; /* seen before, useless */
    }
  }
  /* new reply, adjust credits! */
  if (hostId != NULL) /* if we are the sender, hostId will be NULL */
    coreAPI->changeTrust(hostId, ite->priority);
  prio = ite->priority;
  ite->priority = 0; /* no priority for further replies,
			because we don't get paid for those... */
  if (hostId != NULL)
    noticeContentReceived(hostId); /* tell querymanager that this node
				      is giving us content! */
  IFLOG(LOG_DEBUG,
	hash2hex(&ite->hash,
		 &hex));
#if DEBUG_ROUTING
  LOG(LOG_DEBUG, 
      "DEBUG: indirecting new content matching query %s.\n",
      &hex);
#endif
  sendReply(ite,
	    msg); 
  GROW(ite->seen,
       ite->seenIndex,
       ite->seenIndex+1);
  memcpy(&ite->seen[ite->seenIndex-1],
	 &contentHC,
	 sizeof(HashCode160));
  statChange(stat_content_in_ok, 1);
  MUTEX_UNLOCK(&lock);
  return prio;
}


/* end of routing.c */
