/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file include/applications/afs/encoding/directory.h
 * @brief functions for building directories
 * @author Christian Grothoff
 **/

#ifndef AFS_DIRECTORY_H
#define AFS_DIRECTORY_H

#include "applications/afs/encoding/block.h"

/* what is the context in which a root-node was discovered? */
#define DIR_CONTEXT_SEARCH    1
#define DIR_CONTEXT_INSERT    2
#define DIR_CONTEXT_DIRECTORY 4
#define DIR_CONTEXT_INSERT_SB 8

#define DIR_CONTEXT_ALL (DIR_CONTEXT_SEARCH|DIR_CONTEXT_INSERT|DIR_CONTEXT_DIRECTORY|DIR_CONTEXT_INSERT_SB)

/* see also: http://www.w3.org/TR/PNG#R.PNG-file-signature */
#define GNUNET_DIRECTORY_MAGIC "\211GND\r\n\032\n"
#define GNUNET_DIRECTORY_EXT ".gnd"
#define GNUNET_DIRECTORY_MIME "application/gnunet-directory"

/**
 * Format of a GNUnet directory (both in memory and on the drive).
 **/ 
typedef struct {
  char MAGIC[8];

  /* in network byte order */
  unsigned int version;

  /* number of files in the directory */
  unsigned int number_of_files;

  /* description/filename of the directory */
  char description[MAX_DESC_LEN]; 

  /* must be zero for now */
  char reserved[sizeof(RootNode) - MAX_DESC_LEN - 16]; 

  /* number_of_files root-nodes */
  RootNode contents[0];

} GNUnetDirectory;


/**
 * Makes a root-node available for directory building.
 *
 * This function is called whenever a root-node is encountered.  This
 * can either be because the user inserted a file locally; because we
 * received a search result or because the user retrieved a directory
 * with root-nodes.  From which context the root node was encountered
 * is specified in the context parameters.<p>
 *
 * makeRootNodeAvailable adds the node to the list of files that
 * we can build a directory from later.  The context is used to allow
 * the user to filter on root-node sources.
 * 
 * @param root the file identifier that was encountered
 * @param context the context in which the identifier was encountered (may not be a bitmask)
 **/
void makeRootNodeAvailable(RootNode * root,
			   unsigned int context);

/**
 * Remove all of the root-nodes of a particular type
 * from the directory database.
 *
 * @param context bitmask of the databases that should be emptied.
 **/ 
void emptyDirectoryDatabase(unsigned int contexts);

/**
 * Callback function.
 * @param root a root-node
 * @param closure a closure
 **/
typedef void (*RootNodeCallback)(RootNode * root, 
				 void * closure);

/**
 * Iterate over all entries that match the given context
 * mask.
 *
 * @param contexts context bitmask for the entries to iterate over
 * @param callback function to call on each entry, may be NULL
 * @param closure extra argument to the callback
 * @return number of entries found
 **/
int iterateDirectoryDatabase(unsigned int contexts,
			     RootNodeCallback callback,
			     void * closure);

/**
 * Build a GNUnet directory in memoy.
 * 
 * @param numberOfEntries how many files are in the directory
 * @param name what is the name of the directory
 * @param entries the entries in the directory
 * @return the directory
 **/
GNUnetDirectory * buildDirectory(int numberOfEntries,
				 char * name,
				 RootNode * entries);

/**
 * Write a directory to a file.
 * 
 * @param dir the directory
 * @param fn the filename
 * @return OK on success, SYSERR on error
 **/
int writeGNUnetDirectory(GNUnetDirectory * dir,
			 char * fn);

/**
 * Read a directory from a file.
 * 
 * @param fn the filename
 * @return the directory on success, NULL on error
 **/
GNUnetDirectory * readGNUnetDirectory(char * fn);



#endif
